/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "maya/maya.pmh"

#include "maya/MGlobal.h"
#include "maya/MFnPlugin.h"

#define FE_MN_COMPOUND_TEST		FALSE
#define FE_MN_POPULATE_DEBUG	FALSE
#define FE_MN_CATALOG_DEBUG		FALSE
#define FE_MN_PAINT_DEBUG		FALSE
#define FE_MN_COMPUTE_DEBUG		FALSE

// TODO may sure this is a unique name
static const char* feMayaDefaultOutputSurface = "Output Surface";

namespace fe
{

I32 mayaDiscoverOperators(MObject a_object)
{
	MFnPlugin plugin(a_object,PLUGIN_COMPANY,"1.0","Any");

	MayaContext mayaContext;

#if FE_MN_POPULATE_DEBUG
	feLog("mayaDiscoverOperators master=%p\n",
		mayaContext.master().raw());
#endif

	sp<Registry> spRegistry=mayaContext.master()->registry();
	mayaContext.loadLibraries();

	MString feMenu;
	MGlobal::executeCommand("menu -label FE -p MayaWindow",feMenu);
	MayaNode::ms_feMenu=feMenu.asChar();

	MString feCreateMenu;
	MGlobal::executeCommand(
			("menuItem -sm true -label \"Create Operator\" -p "+
			String(feMenu.asChar())).c_str(),feCreateMenu);

	plugin.addMenuItem("Create Mesh",feMenu.asChar(),
			"createNode","mesh");

	int count=0;

	std::vector<String> available;
	spRegistry->listAvailable("OperatorSurfaceI",available);
	U32 candidates=available.size();
	for(U32 index=0;index<candidates;index++)
	{
#if FE_MN_POPULATE_DEBUG
		feLog("checking %d/%d \"%s\"\n",
				index,candidates,available[index].c_str());
#endif

		sp<OperatorSurfaceI> spOperatorSurfaceI=
				spRegistry->create(available[index]);
		if(!spOperatorSurfaceI.isValid())
		{
			feLog("mayaDiscoverOperators"
					" Component is not an OperatorSurfaceI \"%s\"\n",
					available[index].c_str());
			continue;
		}

		String implementation=available[index];
		char* buffer1=strchr(const_cast<char*>(available[index].c_str()),'.');
		if(buffer1)
		{
			buffer1++;
			const char* buffer2=strchr(buffer1,'.');
			if(buffer2>buffer1)
			{
				implementation.sPrintf("%.*s",int(buffer2-buffer1),buffer1);
			}
			else
			{
				implementation=buffer1;
			}
		}

#if FE_MN_POPULATE_DEBUG
		feLog("mayaDiscoverOperators implementation \"%s\"\n",
				implementation.c_str());
#endif

		const U32 uniqueId=0x80013+count;	//* HACK

		const char* name=strdup(implementation.c_str());
		const char* english=strdup(("FE_"+implementation).c_str());

		fe::MayaNode::preRegister(uniqueId,english,available[index]);
		MStatus status=plugin.registerNode(
				english,
				MTypeId(uniqueId),
				fe::MayaNode::create,
				fe::MayaNode::initialize);
		if(!status)
		{
			status.perror("registerNode");
		}
		fe::MayaNode::postRegister();

		plugin.addMenuItem(
				implementation.c_str(),
				feCreateMenu.asChar(),
				"createNode",
				english);

		count++;
	}

	return count;
}

} // namespace fe

extern "C"
{

//* discovery point for Maya
FE_DL_EXPORT MStatus initializePlugin(MObject a_object)
{
#if FE_MN_POPULATE_DEBUG
	feLog("fexMayaDL ::initializePlugin\n");
#endif

	I32 count=fe::mayaDiscoverOperators(a_object);

#if FE_MN_POPULATE_DEBUG
	feLog("fexMayaDL ::initializePlugin done\n");
#endif

	return MS::kSuccess;
}

FE_DL_EXPORT MStatus uninitializePlugin(MObject a_object)
{
#if FE_MN_POPULATE_DEBUG
	feLog("fexMayaDL ::uninitializePlugin\n");
#endif

	MGlobal::executeCommand(
			("deleteUI -m "+fe::MayaNode::ms_feMenu).c_str());

	MFnPlugin plugin(a_object);

	//* TODO proper
/*
	MStatus status=plugin.deregisterNode(fe::MayaNode::ms_id);
	if(!status)
	{
		status.perror("deregisterNode");
	}
*/

#if FE_MN_POPULATE_DEBUG
	feLog("fexMayaDL ::uninitializePlugin done\n");
#endif

	return MS::kSuccess;
}

//* TODO Is this unneeded in Maya?
#if FALSE
FE_DL_EXPORT void HoudiniDSOInit(UT_DSOInfo& dsoinfo)
{
	//* force Houdini to load with RTLD_GLOBAL (needed for dynamic_cast<>)
	dsoinfo.loadGlobal=true;
}
#endif

}

namespace fe
{

I32	MayaNode::ms_registerLock=0;
String MayaNode::ms_registerName;
String MayaNode::ms_feMenu;
MObject	MayaNode::ms_compoundBehavior;
std::map<String,MayaNode::Registration> MayaNode::ms_mayaRegistry;

MayaNode::MayaNode(void):
	m_paintSize(0)
{
//	feLog("MayaNode::MayaNode master=%p\n",m_mayaContext.master().raw());

	sp<Registry> spRegistry=m_mayaContext.master()->registry();
	m_mayaContext.loadLibraries();

	m_spDrawOutput=spRegistry->create("DrawI.MayaDraw");
//	feLog("MayaNode::MayaNode m_spDrawOutput valid=%d\n",
//			m_spDrawOutput.isValid());

	setupSignaler(m_mayaContext.scope());

#if FE_COUNTED_TRACK
	Counted::registerRegion(this,sizeof(MayaNode),"MayaNode");
#endif

//	feLog("MayaNode::MayaNode done\n");
}

MayaNode::~MayaNode(void)
{
//	feLog("~MayaNode\n");

	clearReferences();

#if FE_COUNTED_TRACK
	Counted::deregisterRegion(this);
#endif
}

void MayaNode::preRegister(U32 a_uniqueId,String a_name,
		String a_implementation)
{
#if FE_MN_POPULATE_DEBUG
	feLog("MayaNode::preRegister\n");
#endif

	FEASSERT(ms_registerLock==0);
	ms_registerLock++;
	ms_registerName=a_name;

	Registration& registration=ms_mayaRegistry[a_name];
	registration.m_autoOutput=FALSE;
	registration.m_uniqueId=a_uniqueId;
	registration.m_implementation=a_implementation;
}

void MayaNode::postRegister(void)
{
#if FE_MN_POPULATE_DEBUG
	feLog("MayaNode::postRegister\n");
#endif

	FEASSERT(ms_registerLock==1);

	ms_registerName="";
	ms_registerLock--;
}

MStatus MayaNode::initialize(void)
{
#if FE_MN_POPULATE_DEBUG
	feLog("MayaNode::initialize %s %d\n",
			ms_registerName.c_str(),ms_registerLock);
#endif

	populateOperator(ms_registerName);

#if FE_MN_POPULATE_DEBUG
	feLog("MayaNode::initialize done\n");
#endif

	return MS::kSuccess;
}

void* MayaNode::create(void)
{
//	feLog("MayaNode::create\n");
	return new MayaNode();
}

void MayaNode::populateOperator(String a_name)
{
	MayaContext mayaContext;

#if FE_MN_POPULATE_DEBUG
	feLog("MayaNode:populateOperator master=%p\n",
		mayaContext.master().raw());
#endif

	sp<Registry> spRegistry=mayaContext.master()->registry();

	Registration& registration=ms_mayaRegistry[a_name];
	String implementation=registration.m_implementation;
	std::map<String,Parameter>& rParameterMap=registration.m_parameterMap;

#if FE_MN_POPULATE_DEBUG
	feLog("registration \"%s\" \"%s\"\n",
			a_name.c_str(),implementation.c_str());
#endif

	sp<OperatorSurfaceI> spOperatorSurfaceI=
			spRegistry->create(implementation);
	if(!spOperatorSurfaceI.isValid())
	{
		feLog("MayaNode:populateOperator"
				" Component is not an OperatorSurfaceI \"%s\"\n",
				implementation.c_str());
		return;
	}

	//* tried to move built-in attributes to a compound
	MStatus status=MS::kSuccess;
#if FE_MN_COMPOUND_TEST
//	MFnAttribute fnCaching;
//	fnCaching.setObject(caching);
//	fnCaching.setHidden(true);

//	MFnAttribute fnState;
//	fnState.setObject(state);
//	fnState.setHidden(true);

	MFnCompoundAttribute fnBehavior;
	ms_compoundBehavior=fnBehavior.create("NodeBehavior","behavior",&status);
	status=fnBehavior.setNiceNameOverride("Node Behavior");
	FEASSERT(status);
//	status=fnBehavior.addChild(caching);
	FEASSERT(status);
//	status=fnBehavior.addChild(state);
	FEASSERT(status);
#endif

	std::vector<String> keys;

	sp<Catalog> spCatalog=spOperatorSurfaceI;
	if(!spCatalog.isValid())
	{
		feLog("MayaNode:populateOperator"
				" Component is not a Catalog \"%s\"\n",
				implementation.c_str());

		//* TODO keep this from crashing later
	}
	else
	{
		spCatalog->catalogKeys(keys);
	}

	U32 number=keys.size();
#if FE_MN_POPULATE_DEBUG
	feLog("MayaNode::populateOperator keys %d\n",number);
#endif

	U32 outputs=0;
	for(U32 m=0;m<number;m++)
	{
		Instance instance;
		if(!spCatalog->catalogLookup(keys[m],instance))
		{
			continue;
		}
		if(instance.is< sp<Component> >())
		{
			char* keyname=strdup(keys[m].c_str());
			String io=spCatalog->catalogOrDefault<String>(keyname,"IO","input");
			if(io=="output")
			{
				outputs++;
			}
		}
	}

	if(!outputs)
	{
#if FE_MN_POPULATE_DEBUG
		feLog("MayaNode::populateOperator no output surface specified\n");
#endif

		//* refresh key list
		keys.clear();
		registration.m_autoOutput=TRUE;

		if(spCatalog.isValid())
		{

			spCatalog->catalog< sp<Component> >(feMayaDefaultOutputSurface);
			spCatalog->catalog<String>(feMayaDefaultOutputSurface,"IO")=
					"output";

			spCatalog->catalogKeys(keys);
			number=keys.size();
		}
	}

	for(U32 m=0;m<number;m++)
	{
#if FE_MN_POPULATE_DEBUG
		feLog("MayaNode::populateOperator key %d: \"%s\"\n",
				m,keys[m].c_str());
#endif

		Instance instance;
		bool success=spCatalog->catalogLookup(keys[m],instance);
		if(!success)
		{
			feLog("MayaNode::populateOperator invalid Instance\n");
			continue;
		}

		char* cleanname=strdup(keys[m].substitute(" ","_").c_str());
		char* keyname=strdup(keys[m].c_str());
		String io=spCatalog->catalogOrDefault<String>(keyname,"IO","input");

		if(instance.is< sp<Component> >())
		{
			//* see meshOpNode.cpp

			String implementation=
					spCatalog->catalogOrDefault<String>(keyname,
					"implementation",
					"SurfaceAccessibleI.SurfaceAccessibleMaya");

			if(implementation=="SurfaceAccessibleI" || implementation==
					"SurfaceAccessibleI.SurfaceAccessibleMaya")
			{
				MFnTypedAttribute typedAttr;
				rParameterMap[keyname].m_object=typedAttr.create(
						cleanname,cleanname,MFnMeshData::kMesh);
				typedAttr.setNiceNameOverride(keyname);
				if(io=="input")
				{
					U32 channel=registration.m_inputChannelArray.size();
					registration.m_inputChannelArray.resize(channel+1);
					registration.m_inputChannelArray[channel]=keyname;

					typedAttr.setStorable(true);
				}
				else
				{
					U32 channel=registration.m_outputChannelArray.size();
					registration.m_outputChannelArray.resize(channel+1);
					registration.m_outputChannelArray[channel]=keyname;

					typedAttr.setStorable(false);
					typedAttr.setWritable(false);
				}
				MStatus status=addAttribute(rParameterMap[keyname].m_object);
			}
		}
		else
		{
			MFnAttribute attribute;

			if(instance.is<String>())
			{
				char* value=strdup(instance.cast<String>().c_str());

				MFnStringData fnStringData;
				MObject defaultString=fnStringData.create(value);
				MFnTypedAttribute stringAttr;
				rParameterMap[keyname].m_object=
						stringAttr.create(cleanname,cleanname,
						MFnData::kString,defaultString);
				attribute.setObject(stringAttr.object());
			}
			else
			{
				MFnNumericAttribute numAttr;

				if(instance.is<bool>())
				{
					bool value=instance.cast<bool>();

					rParameterMap[keyname].m_object=
							numAttr.create(cleanname,cleanname,
							MFnNumericData::kBoolean,value);
				}
				else if(instance.is<Real>())
				{
					Real value=instance.cast<Real>();

					Real min=spCatalog->catalogOrDefault<Real>(keyname,
							"min",0.0);
					Real max=spCatalog->catalogOrDefault<Real>(keyname,
							"max",1.0);
					Real low=spCatalog->catalogOrDefault<Real>(keyname,
							"low",min);
					Real high=spCatalog->catalogOrDefault<Real>(keyname,
							"high",max);

					rParameterMap[keyname].m_object=
							numAttr.create(cleanname,cleanname,
							MFnNumericData::kDouble,value);
					numAttr.setSoftMin(low);
					numAttr.setSoftMax(high);
					numAttr.setMin(min);
					numAttr.setMax(max);

					const BWORD paintable=spCatalog->catalogOrDefault<bool>(
							keyname,"paintable",false);
					if(paintable)
					{
						const String paintableName=String(cleanname)+"Painting";
						const String paintListName=paintableName+"List";
						char* paintname=strdup(paintableName.c_str());
						char* listname=strdup(paintListName.c_str());

#if FALSE
						MFnDoubleArrayData doubleArrayData;
						MObject doubleArrayObject=
								doubleArrayData.create(MDoubleArray(30,1.0));

						MFnTypedAttribute paintAttr;
						MObject objectPaint=
								paintAttr.create(paintname,paintname,
								MFnData::kDoubleArray,doubleArrayObject);
#else
						MFnTypedAttribute paintAttr;
						MObject objectPaint=
								paintAttr.create(paintname,paintname,
								MFnData::kDoubleArray);
#endif
						paintAttr.setHidden(true);
						paintAttr.setStorable(true);
//						paintAttr.setUsesArrayDataBuilder(true);

						MFnCompoundAttribute compoundAttr;
						MObject objectPaintList=
								compoundAttr.create(listname,listname);
						compoundAttr.addChild(objectPaint);
						compoundAttr.setHidden(true);
						compoundAttr.setStorable(true);
//						compoundAttr.setArray(true);
//						compoundAttr.setUsesArrayDataBuilder(true);

						MStatus status=addAttribute(objectPaintList);

						rParameterMap[keyname].m_objectPaint=objectPaint;
						rParameterMap[keyname].m_objectPaintList=
								objectPaintList;

						String command="makePaintable"
								" -attrType doubleArray"
								" "+a_name+" "+paintableName;
#if FE_MN_POPULATE_DEBUG
						feLog("MayaNode::populateOperator MEL \"%s\"\n",
								command.c_str());
#endif

						//* making the call causes paint changes to be ignored
//						MGlobal::executeCommand(MString(command.c_str()));
					}
				}
				else if(instance.is<I32>())
				{
					I32 value=instance.cast<I32>();

					I32 min=spCatalog->catalogOrDefault<I32>(keyname,
							"min",0);
					I32 max=spCatalog->catalogOrDefault<I32>(keyname,
							"max",10);
					I32 low=spCatalog->catalogOrDefault<I32>(keyname,
							"low",min);
					I32 high=spCatalog->catalogOrDefault<I32>(keyname,
							"high",max);

					rParameterMap[keyname].m_object=
							numAttr.create(cleanname,cleanname,
							MFnNumericData::kInt,value);
					numAttr.setSoftMin(low);
					numAttr.setSoftMax(high);
					numAttr.setMin(min);
					numAttr.setMax(max);
				}
				else if(instance.is<SpatialVector>())
				{
					SpatialVector& value=instance.cast<SpatialVector>();

					rParameterMap[keyname].m_object=
							numAttr.create(cleanname,cleanname,
							MFnNumericData::k3Double);
					numAttr.setDefault((double)value[0],(double)value[1],
							(double)value[2]);
				}
				else if(instance.is<Vector4>())
				{
					Vector4& value=instance.cast<Vector4>();

					rParameterMap[keyname].m_object=
							numAttr.create(cleanname,cleanname,
							MFnNumericData::k4Double);
					numAttr.setDefault((double)value[0],(double)value[1],
							(double)value[2],(double)value[3]);
				}

				attribute.setObject(numAttr.object());
			}

			attribute.setNiceNameOverride(keyname);
			if(io=="input")
			{
				attribute.setKeyable(true);
				attribute.setStorable(true);
			}
			else
			{
				attribute.setKeyable(false);
				attribute.setStorable(false);
			}

#if FE_MN_COMPOUND_TEST
			MStatus status=
					fnBehavior.addChild(rParameterMap[keyname].m_object);
#else
			MStatus status=addAttribute(rParameterMap[keyname].m_object);
#endif
		}
	}

#if FE_MN_COMPOUND_TEST
	status=addAttribute(ms_compoundBehavior);
	FEASSERT(status);
#endif

	//* Currently every output is dependent on every input.
	//* We could get more specific.
	std::map<String,Parameter>::const_iterator it=rParameterMap.begin();
	while(it!=rParameterMap.end())
	{
		const String& key=it->first;
		String io=spCatalog->catalogOrDefault<String>(key,"IO","input");
		if(io=="output")
		{
			const Parameter& parameter=it->second;
			const MObject& rParamOutput=parameter.m_object;

			std::map<String,Parameter>::const_iterator it2=
					rParameterMap.begin();
			while(it2!=rParameterMap.end())
			{
				const String& key2=it2->first;
				String io2=
						spCatalog->catalogOrDefault<String>(key2,"IO","input");
				if(io2=="input")
				{
					const Parameter& parameter2=it2->second;
					const MObject& rParamInput=parameter2.m_object;

					MStatus status=attributeAffects(rParamInput,rParamOutput);
					if(!status)
					{
						feLog("MayaNode::populateOperator"
								" attributeAffects %s -> %s failed\n",
								key2.c_str(),key.c_str());
					}

					const MObject& rParamPaint=parameter2.m_objectPaint;
					if(!rParamPaint.isNull())
					{
						status=attributeAffects(rParamPaint,rParamOutput);
						if(!status)
						{
							feLog("MayaNode::populateOperator"
									" attributeAffects %s (Paint)"
									" -> %s failed\n",
									key2.c_str(),key.c_str());
						}
					}

					const MObject& rParamPaintList=parameter2.m_objectPaintList;
					if(!rParamPaintList.isNull())
					{
						status=attributeAffects(rParamPaintList,rParamOutput);
						if(!status)
						{
							feLog("MayaNode::populateOperator"
									" attributeAffects %s (PaintList)"
									" -> %s failed\n",
									key2.c_str(),key.c_str());
						}
					}
				}
				it2++;
			}
		}
		it++;
	}

#if FE_MN_POPULATE_DEBUG
	feLog("MayaNode::populateOperator done\n");
#endif
}

void MayaNode::updateCatalog(const MPlug& rPlug,MDataBlock& rData,
		Real time,BWORD a_doInput)
{
#if FE_MN_CATALOG_DEBUG
	feLog("MayaNode::updateCatalog doInput=%d\n",a_doInput);
#endif

	if(!m_spOperatorSurfaceI.isValid())
	{
		feLog("MayaNode::updateCatalog invalid m_spOperatorSurfaceI\n");
		return;
	}

	sp<Catalog> spCatalog=m_spOperatorSurfaceI;
	if(!spCatalog.isValid())
	{
		feLog("MayaNode:updateCatalog operator is not a Catalog\n");
		return;
	}

	Registration& registration=ms_mayaRegistry[m_name];

	BWORD autocopyInput=FALSE;
	if(registration.m_autoOutput)
	{
#if FE_MN_CATALOG_DEBUG
		feLog("MayaNode::updateCatalog auto-output\n");
#endif

		spCatalog->catalog< sp<Component> >(feMayaDefaultOutputSurface);
		spCatalog->catalog<String>(feMayaDefaultOutputSurface,"IO")="output";
		autocopyInput=TRUE;
	}

	const std::map<String,Parameter>& rParameterMap=registration.m_parameterMap;

	std::map<String,Parameter>::const_iterator it=rParameterMap.begin();
	while(it!=rParameterMap.end())
	{
		const String& key=it->first;
		const Parameter& parameter=it->second;
		const MObject& rObject=parameter.m_object;
		const MObject& rObjectPaint=parameter.m_objectPaint;
		const MObject& rObjectPaintList=parameter.m_objectPaintList;

#if FE_MN_CATALOG_DEBUG
		feLog("MayaNode::updateCatalog \"%s\"\n",key.c_str());
#endif

		Instance instance;
		bool success=spCatalog->catalogLookup(key,instance);
		if(!success)
		{
			feLog("MayaNode::updateCatalog invalid Instance \"%s\"\n",
					key.c_str());
			it++;
			continue;
		}

		String io=spCatalog->catalogOrDefault<String>(key,"IO","input");
		BWORD isInput=(io=="input");
		if(isInput!=a_doInput)
		{
//			feLog("MayaNode::updateCatalog skip \"%s\"\n",io.c_str());
			it++;
			continue;
		}

		if(instance.is<String>())
		{
			MStatus status;
			if(isInput)
			{
				MDataHandle handleIn=rData.inputValue(rObject,&status);
				if(status==MS::kSuccess)
				{
					spCatalog->catalog<String>(key)=
							handleIn.asString().asChar();
				}
			}
			else
			{
				MDataHandle handleOut=rData.outputValue(rObject,&status);
				handleOut.set(MString(spCatalog->catalog<String>(key).c_str()));
			}
		}
		else if(instance.is<bool>())
		{
			MStatus status;
			if(isInput)
			{
				MDataHandle handleIn=rData.inputValue(rObject,&status);
				if(status==MS::kSuccess)
				{
					spCatalog->catalog<bool>(key)=handleIn.asBool();
				}
			}
			else
			{
				MDataHandle handleOut=rData.outputValue(rObject,&status);
				handleOut.set(spCatalog->catalog<bool>(key));
			}
		}
		else if(instance.is<Real>())
		{
			Real& rValue=spCatalog->catalog<Real>(key);

			MStatus status;

			if(isInput)
			{
				MDataHandle handleIn=rData.inputValue(rObject,&status);
				if(status==MS::kSuccess)
				{
					rValue=handleIn.asDouble();
				}
			}
			else
			{
				MDataHandle handleOut=rData.outputValue(rObject,&status);
				handleOut.set(double(rValue));
			}

#if FE_MN_CATALOG_DEBUG
			feLog("MayaNode::updateCatalog now Real %.6G\n",rValue);
#endif

			const BWORD paintable=spCatalog->catalogOrDefault<bool>(
					key,"paintable",false);

#if FE_MN_CATALOG_DEBUG
			feLog("MayaNode::updateCatalog paintable %d\n",paintable);
#endif

			if(paintable)
			{
				String rPaintableName=spCatalog->catalog<String>(
						key,"paintableName");
				if(rPaintableName.empty())
				{
					rPaintableName=key.substitute(" ","_")+"Painting";
				}

				std::vector<Real>& rPainting=
						spCatalog->catalog< std::vector<Real> >(key,"painting");

				if(isInput)
				{
#if FE_MN_PAINT_DEBUG
					feLog("MayaNode::updateCatalog"
							" input painting size %d ---- %s\n",
							m_paintSize,rPlug.name().asChar());
#endif

#if FALSE
					MArrayDataHandle listIn=
							rData.inputArrayValue(rObjectPaintList,&status);
					if(status!=MS::kSuccess)
					{
						feLog("MayaNode::updateCatalog input failed to"
								" get handle to painting list %s\n",
								status.errorString().asChar());
					}

					U32 listCount=listIn.elementCount(&status);
					feLog("MayaNode::updateCatalog input painting"
							" listCount %d\n",listCount);
#else
					MDataHandle listIn=
							rData.inputValue(rObjectPaintList,&status);
					if(status!=MS::kSuccess)
					{
						feLog("MayaNode::updateCatalog input failed to"
								" get handle to painting list %s\n",
								status.errorString().asChar());
					}

					U32 listCount=1;
//					feLog("MayaNode::updateCatalog input painting"
//							" listCount %d\n",listCount);
#endif

					for(U32 m=0;m<listCount;m++)
					{
#if FALSE
						status=listIn.jumpToArrayElement(m);
						if(status!=MS::kSuccess)
						{
							feLog("MayaNode::updateCatalog input painting"
									" failed to jump to %d/%d %s\n",
									m,listCount,status.errorString().asChar());
						}
						const U32 index=listIn.elementIndex();
						MDataHandle elementHandle=listIn.inputValue(&status);
						if(status!=MS::kSuccess)
						{
							feLog("MayaNode::updateCatalog input painting"
									" failed to get handle %d/%d %s\n",
									m,listCount,status.errorString().asChar());
						}

						MDataHandle childHandle=
								elementHandle.child(rObjectPaint);
#else
						MDataHandle childHandle=
								listIn.child(rObjectPaint);
#endif
						MObject childData=childHandle.data();
						if(childData.isNull())
						{
#if FE_MN_PAINT_DEBUG
							feLog("MayaNode::updateCatalog input painting"
									" child data null\n");
#endif

							if(m_paintSize)
							{
#if FE_MN_PAINT_DEBUG
								feLog("MayaNode::updateCatalog input painting"
										" doubleArrayData create %d\n",
										m_paintSize);
#endif
								MFnDoubleArrayData doubleArrayData;
								MObject doubleArrayObject=
										doubleArrayData.create(
										MDoubleArray(m_paintSize,1.0));

								childHandle.set(doubleArrayObject);
								childData=childHandle.data();
							}
						}

						MFnDoubleArrayData doubleArrayData(childData,&status);
						if(status!=MS::kSuccess)
						{
#if FE_MN_PAINT_DEBUG
							feLog("MayaNode::updateCatalog input painting"
									" failed to bind doubleArrayData %s\n",
									status.errorString().asChar());
#endif
						}
						else
						{
							U32 length=doubleArrayData.length(&status);
							if(status!=MS::kSuccess)
							{
								feLog("MayaNode::updateCatalog input painting"
										" failed to get doubleArrayData"
										" length %s\n",
										status.errorString().asChar());
							}
#if FE_MN_PAINT_DEBUG
							feLog("MayaNode::updateCatalog input painting"
									" doubleArrayData length %d\n",length);
#endif

							if(!m_paintSize && length)
							{
#if FE_MN_PAINT_DEBUG
								feLog("MayaNode::updateCatalog input painting"
										" doubleArrayData reload %d\n",length);
#endif
								m_paintSize=length;
							}

							const U32 size=rPainting.size();
							if(size!=length)
							{
								rPainting.resize(length);
							}

							for(U32 m=0;m<length;m++)
							{
								rPainting[m]=doubleArrayData[m];
							}
						}
					}
				}
			}
		}
		else if(instance.is<I32>())
		{
			I32& rValue=spCatalog->catalog<I32>(key);

			MStatus status;
			if(isInput)
			{
				MDataHandle handleIn=rData.inputValue(rObject,&status);
				if(status==MS::kSuccess)
				{
					rValue=handleIn.asInt();
				}
			}
			else
			{
				MDataHandle handleOut=rData.outputValue(rObject,&status);
				handleOut.set(rValue);
			}
#if FE_MN_CATALOG_DEBUG
			feLog("MayaNode::updateCatalog now I32 %d\n",rValue);
#endif
		}
		else if(instance.is<SpatialVector>())
		{
			MStatus status;
			if(isInput)
			{
				MDataHandle handleIn=rData.inputValue(rObject,&status);
				if(status==MS::kSuccess)
				{
					double3& rValue=handleIn.asDouble3();
					set(spCatalog->catalog<SpatialVector>(key),rValue[0],
							rValue[1],rValue[2]);
				}
			}
			else
			{
				SpatialVector& rVector=spCatalog->catalog<SpatialVector>(key);

				MDataHandle handleOut=rData.outputValue(rObject,&status);
				handleOut.set(rVector[0],rVector[1],rVector[2]);
			}
		}
		else if(instance.is<Vector4>())
		{
			// TODO four floats is oddly supported
		}
		else if(instance.is< sp<Component> >())
		{
			//* meshes handled elsewhere
		}

		it++;
	}

	sp<Registry> spRegistry=m_mayaContext.master()->registry();

	std::vector<String> keys;
	spCatalog->catalogKeys(keys);

	const U32 number=keys.size();
	for(U32 m=0;m<number;m++)
	{
		Instance instance;
		bool success=spCatalog->catalogLookup(keys[m],instance);
		if(!success)
		{
			feLog("MayaNode::updateCatalog invalid Instance \"%s\"\n",
					keys[m].c_str());
			continue;
		}

		char* cleanname=strdup(keys[m].substitute(" ","_").c_str());
		char* keyname=strdup(keys[m].c_str());

		if(instance.is< sp<Component> >())
		{
			sp<Component>& rspComponent=instance.cast< sp<Component> >();

			if(!rspComponent.isValid())
			{
				String implementation=spCatalog->catalogOrDefault<String>(
						keyname,"implementation",
						"SurfaceAccessibleI.SurfaceAccessibleMaya");
				if(!implementation.empty())
				{
					if(implementation=="SurfaceAccessibleI")
					{
						implementation=
								"SurfaceAccessibleI.SurfaceAccessibleMaya";
					}
					rspComponent=spRegistry->create(implementation);
				}
				sp<SurfaceAccessibleI> spSurfaceAccessibleI=rspComponent;
				if(spSurfaceAccessibleI.isValid())
				{
					String io=spCatalog->catalogOrDefault<String>(
							keyname,"IO","input");
					if(io=="input")
					{
						I32 channel=m_surfaceInputArray.size();
#if FE_MN_CATALOG_DEBUG
						feLog("MayaNode::updateCatalog"
								" adding surface input %d \"%s\" \"%s\"\n",
								channel,cleanname,keyname);
#endif
						spCatalog->catalog<I32>(keyname,"channel")=channel;
						m_surfaceInputArray.push_back(spSurfaceAccessibleI);

						if(autocopyInput)
						{
							spCatalog->catalog<String>(
									feMayaDefaultOutputSurface,"copy")=keyname;
							autocopyInput=FALSE;
						}
					}
					else
					{
#if FE_MN_CATALOG_DEBUG
						feLog("MayaNode::updateCatalog"
								" adding surface output \"%s\" \"%s\"\n",
								cleanname,keyname);
#endif
						m_spSurfaceOutput=spSurfaceAccessibleI;

						// HACK
						m_nameOutput=keyname;
					}
				}
			}
		}
	}

#if FE_MN_CATALOG_DEBUG
	feLog("MayaNode::updateCatalog done\n");
#endif
}

MStatus MayaNode::compute(const MPlug& rPlug,MDataBlock& rData)
{
#if FE_MN_COMPUTE_DEBUG
	feLog("MayaNode::compute \"%s\"\n",rPlug.name().asChar());
#endif

	// TODO check maya node state (normal, passthrough, etc)

	m_name=typeName().asChar();
	Registration& rRegistration=ms_mayaRegistry[m_name];
	const U32 uniqueId=rRegistration.m_uniqueId;
	m_implementation=rRegistration.m_implementation;

#if FE_MN_COMPUTE_DEBUG
	feLog("  name \"%s\"\n",m_name.c_str());
	feLog("  unique ID %p\n",uniqueId);
	feLog("  implementation \"%s\"\n",m_implementation.c_str());
#endif

	std::map<String,Parameter>& rParameterMap=rRegistration.m_parameterMap;

	double time=0.0;			// TODO

	String operatorName=m_implementation;

	if(!m_spOperatorSurfaceI.isValid())
	{
		sp<Registry> spRegistry=m_mayaContext.master()->registry();
		m_spOperatorSurfaceI=spRegistry->create(operatorName);

#if FE_MN_COMPUTE_DEBUG
		feLog("MayaNode::compute m_spOperatorSurfaceI \"%s\" valid=%d\n",
				operatorName.c_str(),m_spOperatorSurfaceI.isValid());
#endif

		if(m_spOperatorSurfaceI.isValid())
		{
			updateCatalog(rPlug,rData,time,TRUE);

			m_lastName=operatorName;
			insertHandler(m_spOperatorSurfaceI);
		}
	}
	else
	{
		updateCatalog(rPlug,rData,time,TRUE);
	}

	if(m_spOperatorSurfaceI.isValid())
	{
		double frame=0;	// TODO

		sp<SurfaceAccessibleMaya> spSurfaceOutput=m_spSurfaceOutput;
		if(!spSurfaceOutput.isValid())
		{
			sp<Registry> spRegistry=m_mayaContext.master()->registry();
			spSurfaceOutput=spRegistry->create(
					"SurfaceAccessibleI.SurfaceAccessibleMaya");
		}
#if FE_MN_COMPUTE_DEBUG
		feLog("MayaNode::compute spSurfaceOutput valid=%d\n",
				spSurfaceOutput.isValid());
#endif

		I32 channel=0;
		if(!m_nameOutput.empty())
		{
			sp<Catalog> spCatalog=m_spOperatorSurfaceI;
			String copy=spCatalog->catalogOrDefault<String>(m_nameOutput,
					"copy","");
			channel= -1;
			if(!copy.empty())
			{
				channel=spCatalog->catalogOrDefault<I32>(copy,"channel",-1);
			}
		}

		// TODO multiple outputs with individual channels?

		String outputName=rRegistration.m_outputChannelArray[0];
		MObject paramOutMesh=rParameterMap[outputName].m_object;

		MStatus status=MS::kSuccess;
		MDataHandle outputData=rData.outputValue(paramOutMesh,&status);
		FEASSERT(status);
		MObject meshOutput=outputData.asMesh();

		MDataHandle channelData;

		U32 surfaces=m_surfaceInputArray.size();
		FEASSERT(surfaces==rRegistration.m_inputChannelArray.size());
		for(int surface=0;surface<surfaces;surface++)
		{
			sp<SurfaceAccessibleMaya> spSurfaceInput=
					m_surfaceInputArray[surface];

			String inputName=rRegistration.m_inputChannelArray[surface];
			MObject paramInMesh=rParameterMap[inputName].m_object;

			MStatus status=MS::kSuccess;

			MDataHandle inputData=rData.inputValue(paramInMesh,&status);
			FEASSERT(status);
			MObject meshInput=inputData.asMesh();

			spSurfaceInput->setMeshData(inputData);

			if(surface==channel)
			{
				channelData=inputData;
			}

#if FALSE
			// HACK alternate attribute test
			sp<SurfaceAccessorI> spAccessor=spSurfaceInput->accessor(
					SurfaceAccessibleI::e_point,"milliIndex");
			FEASSERT(spAccessor.isValid());

			if(spAccessor.isValid())
			{
				const I32 pointCount=spAccessor->count();
				for(I32 index=0;index<pointCount;index++)
				{
//					feLog("milli %d/%d\n",index,spAccessor->count());
					spAccessor->set(index,0,Real(0.001*index));
				}
			}
#endif
		}

		if(channel>=0 && channel<surfaces)
		{
#if FE_MN_COMPUTE_DEBUG
			feLog("MayaNode::compute output=input\n");
#endif

			MFnMeshData meshDataFn;
			MObject newMeshData=meshDataFn.create();

			MFnMesh newMeshFn;
			newMeshFn.copy(channelData.asMesh(),newMeshData,&status);
			if(!status)
			{
				feLog("MayaNode::compute mesh copy failed: %s\n",
						status.errorString().asChar());
				return status;
			}

			outputData.set(newMeshData);
			meshOutput=outputData.asMesh();

			const U32 vertices=newMeshFn.numVertices();
			if(m_paintSize!=vertices)
			{
				m_paintSize=vertices;

				//* second update
				updateCatalog(rPlug,rData,time,TRUE);
			}
		}
		else
		{
			// erase all prior geometry

#if FE_MN_COMPUTE_DEBUG
			feLog("MayaNode::compute output reset\n");
#endif

			MFnMeshData meshDataFn;
			MObject newMeshData=meshDataFn.create();

			int numVertices=0;
			int numPolygons=0;
			MFloatPointArray vertexArray;
			MIntArray polygonCounts;
			MIntArray polygonConnects;

			status=MS::kSuccess;
			MFnMesh newMeshFn;
			MObject mesh=newMeshFn.create(numVertices,numPolygons,
					vertexArray,polygonCounts,polygonConnects,
					newMeshData,&status);
			if(!status)
			{
				feLog("MayaNode::compute create failed: %s\n",
						status.errorString().asChar());
			}

			outputData.set(newMeshData);
			meshOutput=outputData.asMesh();
			m_paintSize=0;
		}
		spSurfaceOutput->setMeshData(outputData);

		sp<MayaDraw> spMayaDraw=m_spDrawOutput;
		if(spMayaDraw.isValid())
		{
#if FE_MN_COMPUTE_DEBUG
			feLog("MayaNode::compute set draw mesh\n");
#endif
			spMayaDraw->setMeshObject(meshOutput);
		}

		m_aSurfaceOutput(m_cookSignal)=spSurfaceOutput;
		m_spSignalerI->signal(m_cookSignal);

		updateCatalog(rPlug,rData,time,FALSE);
	}

	m_lastTime=time;

#if FE_MN_COMPUTE_DEBUG
	feLog("MayaNode::compute done\n");
#endif

	return MS::kSuccess;
}

} // namespace
