/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_BDO_DEBUG		FALSE
#define FE_BDO_IO_DEBUG		FALSE
#define FE_BDO_KEY_DEBUG	FALSE

#include "operator/operator.pmh"

using namespace fe;

void BenderOp::initParam(String a_prefix,String a_label,BWORD a_joined)
{
	const String key=a_prefix+a_label;
	catalog<Real>(key)=0.0;
	catalog<Real>(key,"low")= -2.0;
	catalog<Real>(key,"high")=2.0;
	catalog<String>(key,"IO")="output";
	catalog<String>(key,"label")=a_label;
	catalog<String>(key,"page")="Edit";
	catalog<bool>(key,"joined")=a_joined;
}

void BenderOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	for(U32 m=0;m<2;m++)
	{
		const String prefix=m? "after": "before";

		const String key=prefix+"Frame";
		const String label=String(m? "Following": "Leading")+" Frame";
		catalog<Real>(key)=0.0;
		catalog<Real>(key,"high")=1e3;
		catalog<Real>(key,"max")=1e6;
		catalog<String>(key,"IO")="output";
		catalog<String>(key,"label")=label;
		catalog<String>(key,"page")="Edit";

		initParam(prefix,"SpinX",TRUE);
		initParam(prefix,"SpinZ",FALSE);
		initParam(prefix,"TwistX",TRUE);
		initParam(prefix,"TwistY",FALSE);
	}

	catalog<bool>("fastPower")=true;
	catalog<String>("fastPower","label")="Fast Power";
	catalog<String>("fastPower","page")="Behavior";

	catalog<Real>("linearity")=0.45;
	catalog<Real>("linearity","max")=1.0;
	catalog<String>("linearity","label")="Linearity";
	catalog<String>("linearity","enabler")="fastPower";
	catalog<String>("linearity","page")="Behavior";

	catalog< sp<Component> >("Brush");
	catalog<String>("Brush","implementation")="DrawI";
	catalog<String>("Brush","prompt")="LMB picks and bends a primitive."
			"  MMB twists the primitive."
			"  DEL removes a key (shift for all its keys).";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Driver Surface");
	catalog<bool>("Driver Surface","optional")=true;

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
	catalog<bool>("Output Surface","recycle")=true;
}

void BenderOp::handle(Record& a_rSignal)
{
	m_fastPower=catalog<bool>("fastPower");
	m_linearity=catalog<Real>("linearity");

	sp<DrawI> spDrawBrush;
	accessBrush(spDrawBrush,a_rSignal,e_quiet);

#if FE_BDO_DEBUG
	feLog("BenderOp::handle brush %d brushed %d\n",
			spDrawBrush.isValid(),m_brushed);
#endif

	if(spDrawBrush.isValid())
	{
		m_brushed=TRUE;

		const Color white(1.0,1.0,0.8,1.0);
		const Color grey(0.8,0.8,1.0,1.0);
		const Color red(0.8,0.0,0.0,1.0);
		const Color green(0.0,0.5,0.0,1.0);
		const Color blue(0.0,0.0,1.0,1.0);
		const Color cyan(0.0,0.8,1.0,1.0);
		const Color magenta(0.8,0.0,0.8,1.0);
		const Color yellow(0.8,0.8,0.0,1.0);

		const SpatialVector up(0.0,1.0,0.0);

		const Real radius=0.1;
		const SpatialVector scale(radius,radius,radius);

		if(m_spOutput.isValid())
		{
			const SpatialVector& rRayOrigin=rayOrigin(a_rSignal);
			const SpatialVector& rRayDirection=rayDirection(a_rSignal);

			m_event.bind(windowEvent(a_rSignal));

#if FE_BDO_DEBUG
			feLog("%s\n",c_print(m_event));
#endif

			if(m_event.isKeyPress(WindowEvent::e_itemAny))
			{
				U32 item=m_event.item();
				BWORD shifted=FALSE;
				if(item&WindowEvent::e_keyShift)
				{
					shifted=TRUE;
					item^=WindowEvent::e_keyShift;
				}

				if(item==WindowEvent::e_keyDelete)
				{
					if(m_picked>=0)
					{
						if(shifted)
						{
							removeChannel(m_picked);
						}
						else
						{
							removeKey(m_picked,m_frame);
						}
						m_key=Record();
					}
				}
			}

			const WindowEvent::MouseButtons buttons=m_event.mouseButtons();
			if(buttons)
			{
				const I32 mouseX=m_event.mouseX();
				const I32 mouseY=m_event.mouseY();

				const I32 moveX=mouseX-m_lastX;
				const I32 moveY=mouseY-m_lastY;

				SpatialVector groundRay=rRayDirection;
				groundRay[1]=0.0;
				normalize(groundRay);
				const Real deltaX=groundRay[2]*moveY+groundRay[0]*moveX;
				const Real deltaZ=groundRay[0]*moveY-groundRay[2]*moveX;

				if(buttons&WindowEvent::e_mbLeft)
				{
					if(m_event.isMouseDrag())
					{
						if(!m_key.isValid())
						{
							m_key=keyAt(m_picked,m_frame);
						}
						if(m_key.isValid())
						{
							m_aSpinX(m_key)+=0.01*deltaX;
							m_aSpinZ(m_key)+=0.01*deltaZ;
						}
					}
					else
					{
						const Real maxDistance=100.0;
						sp<SurfaceI::ImpactI> spImpact=m_spOutput->rayImpact(
								rRayOrigin,rRayDirection,maxDistance);
						m_picked= -1;
						if(spImpact.isValid())
						{
							sp<SurfaceTriangles::Impact> spTriImpact=spImpact;
							if(spTriImpact.isValid())
							{
								const I32 face=spTriImpact->face();
								const I32 pickIndex=spTriImpact->pickIndex();
#if FE_BDO_DEBUG
								feLog("BenderOp::handle face %d pick %d\n",
										face,pickIndex);
#endif
								m_picked=pickIndex;
							}
						}
						m_key=Record();
					}
				}

				if(buttons&WindowEvent::e_mbMiddle)
				{
					if(m_event.isMouseDrag())
					{
						if(!m_key.isValid())
						{
							m_key=keyAt(m_picked,m_frame);
						}
						if(m_key.isValid())
						{
							m_aTwistX(m_key)+=0.01*moveX;
							m_aTwistY(m_key)+=0.01*moveY;
						}
					}
				}

				m_lastX=mouseX;
				m_lastY=mouseY;
			}

			BWORD drewPicked=FALSE;
			const U32 count=m_spOutputVertices->count();

			sp<RecordArray> spRA=m_spChannels->getArray(m_spChannel);
			for(int i=0;i<spRA->length();i++)
			{
				const I32 primIndex=m_aIndex(spRA,i);
				if(primIndex<0 || primIndex>=count)
				{
					feLog("BenderOp::handle bad primIndex index %d/%d\n",
							primIndex,count);
					removeChannel(primIndex);
					i--;
					continue;
				}

				Record channel=spRA->getRecord(i);
				Record key=evalAt(channel,m_frame);

				const U32 subCount=m_spOutputVertices->subCount(primIndex);
				if(subCount)
				{
					SpatialTransform root;
					SpatialTransform end;
					BWORD success=calculateExtent(m_spInputVertices,primIndex,
							root,end);

					SpatialTransform base;
					SpatialTransform diff;
					calculateTransforms(key,root,base,diff);

					SpatialVector line[4];

					const BWORD picked=(primIndex==m_picked);
					if(picked)
					{
						line[0]=base.translation();
						line[1]=(base*diff).translation();
						line[2]=line[1];
						line[2][1]=0.0;
						line[3]=line[0];
						spDrawBrush->drawLines(line,NULL,4,DrawI::e_strip,
								false,&magenta);
						drewPicked=TRUE;
					}

					set(line[0]);
					line[1]=SpatialVector(2.0,0.0,0.0);

					const U32 rings=picked? 16: 8;
					const Color* pRingColor=picked? &blue: &grey;
					for(U32 m=0;m<rings;m++)
					{
						SpatialTransform between;
						interpolate(between,base,diff,m/Real(rings-1));

						if(picked)
						{
							spDrawBrush->drawTransformedLines(between,&scale,
									line,NULL,2,DrawI::e_discrete,false,&red);
						}

						rotate(between,90.0*degToRad,e_xAxis);
						spDrawBrush->drawCircle(between,&scale,*pRingColor);
					}
				}
			}
			if(!drewPicked && m_picked>=0)
			{
				const SpatialVector xAxis(1.0,0.0,0.0);
				const SpatialVector yAxis(0.0,1.0,0.0);
				const SpatialVector rootPoint=
						m_spOutputVertices->spatialVector(m_picked,0);
				SpatialTransform base;
				makeFrameNormalY(base,rootPoint,xAxis,yAxis);
				rotate(base,90.0*degToRad,e_xAxis);
				spDrawBrush->drawCircle(base,&scale,green);
			}
		}

		return;
	}

	m_frame=currentFrame(a_rSignal);

	if(!access(m_spInputVertices,"Input Surface",
			e_primitive,e_vertices)) return;

	if(!accessOutput(m_spOutputVertices,a_rSignal,
			e_primitive,e_vertices)) return;

	if(!accessOutput(m_spOutput,a_rSignal)) return;

	access(m_spDriver,"Driver Surface",e_quiet);

	const BWORD paramChanged=!m_brushed;
	m_brushed=FALSE;

	const BWORD rebuild=(paramChanged ||
			catalog<bool>("Output Surface","replaced"));

	const U32 count=m_spInputVertices->count();
	const BWORD hasPicked=(m_picked>=0 && m_picked<count);
	BWORD refreshPicked=hasPicked;

	sp<RecordArray> spRA=m_spChannels->getArray(m_spChannel);
	for(int i=0;i<spRA->length();i++)
	{
		const I32 primIndex=m_aIndex(spRA,i);
		if(primIndex==m_picked)
		{
			refreshPicked=FALSE;
		}
		else if(!rebuild)
		{
			//* recycled data should be valid
			continue;
		}

		if(primIndex<0 || primIndex>=count)
		{
			feLog("BenderOp::handle bad primIndex index %d/%d\n",
					primIndex,count);
			removeChannel(primIndex);
			i--;
			continue;
		}

		Record channel=spRA->getRecord(i);
		Record key=evalAt(channel,m_frame);

		const U32 subCount=m_spInputVertices->subCount(primIndex);
		if(subCount)
		{
			SpatialTransform root;
			SpatialTransform end;
			BWORD success=calculateExtent(m_spInputVertices,primIndex,
					root,end);

			SpatialTransform base;
			SpatialTransform diff;
			calculateTransforms(key,root,base,diff);

			SpatialTransform invRoot;
			invert(invRoot,root);

			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const SpatialVector point=m_spInputVertices->spatialVector(
						primIndex,subIndex);

				SpatialVector local;
				transformVector(invRoot,point,local);

				SpatialTransform between;
				interpolate(between,base,diff,local[1]);
				local[1]=0.0;

				SpatialVector transformed;
				transformVector(between,local,transformed);

				m_spOutputVertices->set(primIndex,subIndex,transformed);
			}
		}
	}

	if(refreshPicked)
	{
		const U32 subCount=m_spInputVertices->subCount(m_picked);
		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const SpatialVector point=m_spInputVertices->spatialVector(
					m_picked,subIndex);

			m_spOutputVertices->set(m_picked,subIndex,point);
		}
	}

	if(hasPicked)
	{
		Record before;
		Record after;
		findKeys(m_picked,m_frame,before,after);

		if(before.isValid())
		{
			catalog<Real>("beforeFrame")=m_aFrame(before);
			catalog<Real>("beforeSpinX")=m_aSpinX(before);
			catalog<Real>("beforeSpinZ")=m_aSpinZ(before);
			catalog<Real>("beforeTwistX")=m_aTwistX(before);
			catalog<Real>("beforeTwistY")=m_aTwistY(before);
		}
		else
		{
			catalog<Real>("beforeFrame")=0.0;
			catalog<Real>("beforeSpinX")=0.0;
			catalog<Real>("beforeSpinZ")=0.0;
			catalog<Real>("beforeTwistX")=0.0;
			catalog<Real>("beforeTwistY")=0.0;
		}

		if(after.isValid())
		{
			catalog<Real>("afterFrame")=m_aFrame(after);
			catalog<Real>("afterSpinX")=m_aSpinX(after);
			catalog<Real>("afterSpinZ")=m_aSpinZ(after);
			catalog<Real>("afterTwistX")=m_aTwistX(after);
			catalog<Real>("afterTwistY")=m_aTwistY(after);
		}
		else
		{
			catalog<Real>("afterFrame")=0.0;
			catalog<Real>("afterSpinX")=0.0;
			catalog<Real>("afterSpinZ")=0.0;
			catalog<Real>("afterTwistX")=0.0;
			catalog<Real>("afterTwistY")=0.0;
		}
	}

	return;
}

BWORD BenderOp::calculateExtent(sp<SurfaceAccessorI>& a_rspVertices,
	U32 a_primIndex,SpatialTransform& a_rRoot,SpatialTransform& a_rEnd)
{
	const U32 subCount=a_rspVertices->subCount(a_primIndex);
	if(!subCount)
	{
		return FALSE;
	}

	const SpatialVector rootPoint=
			a_rspVertices->spatialVector(a_primIndex,0);

	Real maxDist2=0.0;
	SpatialVector endPoint=rootPoint;

	for(U32 subIndex=0;subIndex<subCount;subIndex++)
	{
		const SpatialVector point=a_rspVertices->spatialVector(
				a_primIndex,subIndex);

		const Real dist2=magnitudeSquared(point-rootPoint);
		if(maxDist2<dist2)
		{
			maxDist2=dist2;
			endPoint=point;
		}
	}

	const SpatialVector from(0.0,1.0,0.0);
	const SpatialVector to=unitSafe(endPoint-rootPoint);

	set(a_rRoot,from,to);
	a_rEnd=a_rRoot;

	setTranslation(a_rRoot,rootPoint);
	setTranslation(a_rEnd,endPoint);
}

void BenderOp::calculateTransforms(Record key,
	const SpatialTransform& a_rRoot,
	SpatialTransform& a_rBase,SpatialTransform& a_rDiff)
{
	FEASSERT(key.isValid());
	Real& rSpinX=m_aSpinX(key);
	Real& rSpinZ=m_aSpinZ(key);
	Real& rTwistX=m_aTwistX(key);
	Real& rTwistY=m_aTwistY(key);

	const Real maxAngle=0.48*fe::pi-rTwistY;
	Real spin=sqrt(rSpinX*rSpinX+rSpinZ*rSpinZ);
	if(spin>maxAngle)
	{
		const Real rescale=maxAngle/spin;
		rSpinX*=rescale;
		rSpinZ*=rescale;
		spin=maxAngle;
	}

	const SpatialVector axis=unitSafe(SpatialVector(rSpinX,0.0,-rSpinZ));

	SpatialQuaternion tilt(spin,axis);

	const SpatialVector zero(0.0,0.0,0.0);
	const SpatialVector xAxis(1.0,0.0,0.0);
	const SpatialVector yAxis(0.0,1.0,0.0);
	SpatialVector target;
	rotateVector(tilt,yAxis,target);
//	target+=a_rRoot.translation();

	SpatialTransform twist;
	setIdentity(twist);
	rotate(twist,-rTwistX,e_yAxis);

	SpatialQuaternion baseBend(-rTwistY,axis);
	SpatialVector baseTan;
	rotateVector(baseBend,xAxis,baseTan);
	SpatialVector baseDir;
	rotateVector(baseBend,yAxis,baseDir);

	SpatialQuaternion tipBend(spin*2.0+rTwistY,axis);
	SpatialVector tipTan;
	rotateVector(tipBend,xAxis,tipTan);
	SpatialVector tipDir;
	rotateVector(tipBend,yAxis,tipDir);

//	makeFrameNormalY(a_rBase,a_rRoot.translation(),baseTan,baseDir);
	makeFrameNormalY(a_rBase,zero,baseTan,baseDir);

	SpatialTransform tip;
	makeFrameNormalY(tip,target,tipTan,tipDir);

	a_rBase=a_rBase*a_rRoot;
	tip=tip*a_rRoot;

	SpatialTransform invBase;
	invert(invBase,a_rBase);

	const SpatialTransform diff=invBase*twist*tip;

	//* HACK shrink with curvature
	if(!m_fastPower)
	{
		m_matrixPower.solve(a_rDiff,diff,cos(0.5*spin));
	}
	else
	{
		SpatialTransform partial;
		interpolate(partial,a_rBase,diff,cos(0.5*spin));
		a_rDiff=invBase*partial;
	}

#if FE_BDO_DEBUG
	feLog("base\n%s\n",c_print(base));
	feLog("tip\n%s\n",c_print(tip));
#endif
}

void BenderOp::interpolate(SpatialTransform& a_rPartial,
	const SpatialTransform& a_rBase,
	const SpatialTransform& a_rDiff, Real a_power)
{
	if(!m_fastPower)
	{
		SpatialTransform partialDiff;
		m_matrixPower.solve(partialDiff,a_rDiff,a_power);
		a_rPartial=a_rBase*partialDiff;
		return;
	}

	const SpatialVector rootPoint=a_rBase.translation();
	const SpatialTransform tip=a_rBase*a_rDiff;

	SpatialTransform localBase=a_rBase;
	SpatialTransform localTip=tip;
	localBase.translation()-=rootPoint;
	localTip.translation()-=rootPoint;

	SpatialTransform invLocalBase;
	invert(invLocalBase,localBase);

	const SpatialTransform localFull=invLocalBase*localTip;

	const Real power1=1.0-a_power;
	const Real distance=magnitude(localFull.translation());

	const Real arm=0.01;

	const Real yRef=distance*a_power;
	const SpatialVector locRef(0.0,yRef,0.0);
	const SpatialVector xArmRef(arm,yRef,0.0);
	const SpatialVector yArmRef(0.0,yRef+arm,0.0);

	const Real yFull= -distance*power1;
	const SpatialVector locFull(0.0,yFull,0.0);
	const SpatialVector xArmFull(arm,yFull,0.0);
	const SpatialVector yArmFull(0.0,yFull+arm,0.0);

	SpatialVector locDef;
	SpatialVector xArmDef;
	SpatialVector yArmDef;

	transformVector(localFull,locFull,locDef);
	transformVector(localFull,xArmFull,xArmDef);
	transformVector(localFull,yArmFull,yArmDef);

	const SpatialVector locBlend=locRef*power1+locDef*a_power;
	const SpatialVector xArmBlend=xArmRef*power1+xArmDef*a_power;
	const SpatialVector yArmBlend=yArmRef*power1+yArmDef*a_power;

	const SpatialVector locLinear=
			localBase.translation()*power1+localTip.translation()*a_power;
	const SpatialVector locMix=
			locLinear*m_linearity+locBlend*(1.0-m_linearity);

	SpatialTransform localPart;
	makeFrameNormalY(localPart,locMix,
			unitSafe(xArmBlend-locBlend),unitSafe(yArmBlend-locBlend));

	SpatialTransform newTip=localBase*localPart;
	newTip.translation()+=rootPoint;

	a_rPartial=newTip;
}

void BenderOp::setupState(void)
{
	m_spState=registry()->create("Scope");
	m_spState->setName("BenderOp.State");

	m_aFrame.setup(m_spState,"Frame");
	m_aSpinX.setup(m_spState,"SpinX");
	m_aSpinZ.setup(m_spState,"SpinZ");
	m_aTwistX.setup(m_spState,"TwistX");
	m_aTwistY.setup(m_spState,"TwistY");

	m_spKey=m_spState->declare("Key");
	m_spKey->populate(m_aFrame);
	m_spKey->populate(m_aSpinX);
	m_spKey->populate(m_aSpinZ);
	m_spKey->populate(m_aTwistX);
	m_spKey->populate(m_aTwistY);

	m_aIndex.setup(m_spState,"Index");
	m_aKeys.setup(m_spState,"Keys");

	m_spChannel=m_spState->declare("Channel");
	m_spChannel->populate(m_aIndex);
	m_spChannel->populate(m_aKeys);

	m_spChannels=new RecordGroup();
}

BWORD BenderOp::saveState(String& rBuffer)
{
#if FE_BDO_IO_DEBUG
	feLog("BenderOp::saveState\n");
#endif

	sp<data::StreamI> spStream(new data::AsciiStream(m_spState));
	std::ostringstream streamOut;
	spStream->output(streamOut,m_spChannels);

	std::stringbuf* pStringBuf=streamOut.rdbuf();
	rBuffer=pStringBuf->str().c_str();

#if FE_BDO_IO_DEBUG
	feLog("---------\n");
	feLog("%s",rBuffer.c_str());
	feLog("---------\n");
#endif

	return TRUE;
}

BWORD BenderOp::loadState(const String& rBuffer)
{
#if FE_BDO_IO_DEBUG
	feLog("BenderOp::loadState\n");
	feLog("---------\n");
	feLog("%s",rBuffer.c_str());
	feLog("---------\n");
#endif

	setupState();

	std::string stringIn=rBuffer.c_str();
	std::istringstream streamIn(stringIn);

	sp<data::StreamI> spStream(new data::AsciiStream(m_spState));
	sp<RecordGroup> spLoadedGroup=spStream->input(streamIn);

	if(spLoadedGroup.isValid())
	{
		m_spChannels=spLoadedGroup;
	}

	m_spChannelMap=m_spChannels->lookupMap(m_aIndex);

	return TRUE;
}

Record BenderOp::removeChannel(I32 a_index)
{
	Record channel=m_spChannelMap->lookup(a_index);

	return removeChannel(channel);
}

Record BenderOp::removeChannel(Record a_channel)
{
	if(a_channel.isValid())
	{
#if FE_BDO_KEY_DEBUG
		feLog("remove channel %d\n",m_aIndex(a_channel));
#endif

		m_spChannels->remove(a_channel);
		return a_channel;
	}

	return Record();
}

Record BenderOp::keyAt(I32 a_index,Real a_frame)
{
	Record channel=m_spChannelMap->lookup(a_index);

	if(!channel.isValid())
	{
#if FE_BDO_KEY_DEBUG
		feLog("create channel %d\n",a_index);
#endif

		channel=m_spState->createRecord(m_spChannel);
		m_aIndex(channel)=a_index;

		m_spChannels->add(channel);
	}

	return keyAt(channel,a_frame);
}

Record BenderOp::keyAt(Record a_channel,Real a_frame)
{
	const Real threshold=1e-3;

	Record key;

	sp<RecordArray> spRA=m_aKeys(a_channel);
	if(!spRA.isValid())
	{
		spRA=m_spState->createRecordArray(m_spKey,0);
		m_aKeys(a_channel)=spRA;
	}
	for(int i=0;i<spRA->length();i++)
	{
		if(fabs(m_aFrame(spRA,i)-a_frame)<threshold)
		{
			key=spRA->getRecord(i);
			break;
		}
	}

	if(!key.isValid())
	{
		key=evalAt(a_channel,a_frame);

		if(!key.isValid())
		{
#if FE_BDO_KEY_DEBUG
			feLog("create key %d %.6G\n",m_aIndex(a_channel),a_frame);
#endif

			key=m_spState->createRecord(m_spKey);
			m_aFrame(key)=a_frame;
			m_aSpinX(key)=0.0;
			m_aSpinZ(key)=0.0;
			m_aTwistX(key)=0.0;
			m_aTwistY(key)=0.0;
		}

		spRA->add(key);
	}

	return key;
}

Record BenderOp::removeKey(I32 a_index,Real a_frame)
{
	Record channel=m_spChannelMap->lookup(a_index);

	if(!channel.isValid())
	{
		return Record();
	}

	return removeKey(channel,a_frame);
}

Record BenderOp::removeKey(Record a_channel,Real a_frame)
{
	const Real threshold=1e-3;

	sp<RecordArray> spRA=m_aKeys(a_channel);
	if(!spRA.isValid())
	{
		return Record();
	}

	Record key;

	for(int i=0;i<spRA->length();i++)
	{
		if(fabs(m_aFrame(spRA,i)-a_frame)<threshold)
		{
#if FE_BDO_KEY_DEBUG
			feLog("remove key %d %.6G\n",m_aIndex(a_channel),a_frame);
#endif

			key=spRA->getRecord(i);
			spRA->remove(key);
			break;
		}
	}

	if(!spRA->length())
	{
		removeChannel(a_channel);
	}

	return key;
}

void BenderOp::findKeys(I32 a_index,Real a_frame,
	Record& a_rBefore,Record& a_rAfter)
{
	Record channel=m_spChannelMap->lookup(a_index);

	if(!channel.isValid())
	{
		return;
	}

	findKeys(channel,a_frame,a_rBefore,a_rAfter);
}

void BenderOp::findKeys(Record a_channel,Real a_frame,
	Record& a_rBefore,Record& a_rAfter)
{
	sp<RecordArray> spRA=m_aKeys(a_channel);
	if(!spRA.isValid())
	{
		return;
	}

	Real closestBefore= -1.0;
	Real closestAfter= -1.0;

	for(int i=0;i<spRA->length();i++)
	{
		const Real otherFrame=m_aFrame(spRA,i);
		if(otherFrame<=a_frame)
		{
			if(closestBefore<0.0 || closestBefore<otherFrame)
			{
				closestBefore=otherFrame;
				a_rBefore=spRA->getRecord(i);
			}
		}
		else
		{
			if(closestAfter<0.0 || closestAfter>otherFrame)
			{
				closestAfter=otherFrame;
				a_rAfter=spRA->getRecord(i);
			}
		}
	}
}

Record BenderOp::evalAt(I32 a_index,Real a_frame)
{
	Record channel=m_spChannelMap->lookup(a_index);

	if(!channel.isValid())
	{
		return Record();
	}

	return evalAt(channel,a_frame);
}

Record BenderOp::evalAt(Record a_channel,Real a_frame)
{
	Record before;
	Record after;
	findKeys(a_channel,a_frame,before,after);

	const Real closestBefore=before.isValid()? m_aFrame(before): -1.0;
	const Real closestAfter=after.isValid()? m_aFrame(after): -1.0;

	if(!before.isValid())
	{
#if FE_BDO_KEY_DEBUG
		feLog("after key %d %.6G\n",m_aIndex(a_channel),closestAfter);
#endif
		Record key=after.clone();
		m_aFrame(key)=a_frame;
		return key;
	}
	if(!after.isValid())
	{
#if FE_BDO_KEY_DEBUG
		feLog("before key %d %.6G\n",m_aIndex(a_channel),closestBefore);
#endif
		Record key=before.clone();
		m_aFrame(key)=a_frame;
		return key;
	}

	const Real ratio=(a_frame-closestBefore)/(closestAfter-closestBefore);
	const Real ratio1=1.0-ratio;

#if FE_BDO_KEY_DEBUG
	feLog("temp key %d %.6G between %.6G %.6G ratio %.6G\n",
			m_aIndex(a_channel),a_frame,closestBefore,closestAfter,ratio);
#endif

	//* temporary tween key
	Record key=before.clone();
	m_aFrame(key)=a_frame;

	//* TEMP for now, just linear
	m_aSpinX(key)=m_aSpinX(before)*ratio1+m_aSpinX(after)*ratio;
	m_aSpinZ(key)=m_aSpinZ(before)*ratio1+m_aSpinZ(after)*ratio;
	m_aTwistX(key)=m_aTwistX(before)*ratio1+m_aTwistX(after)*ratio;
	m_aTwistY(key)=m_aTwistY(before)*ratio1+m_aTwistY(after)*ratio;

	return key;
}
