/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "surface.pmh"

#define	FE_SS_DEBUG				FALSE
#define	FE_SS_TREE_HIT_DEBUG	FALSE

namespace fe
{

SurfaceSearchable::SurfaceSearchable(void):
	m_refinement(0),
	m_vertices(0),
	m_pVertexArray(NULL),
	m_pNormalArray(NULL),
	m_pPickIndexArray(NULL)
{
}

SurfaceSearchable::~SurfaceSearchable(void)
{
	deallocate(m_pPickIndexArray);
	deallocate(m_pNormalArray);
	deallocate(m_pVertexArray);
}

Protectable* SurfaceSearchable::clone(Protectable* pInstance)
{
#if	FE_SS_DEBUG
	feLog("SurfaceSearchable::clone\n");
#endif

	SurfaceSearchable* pSurfaceSearchable= pInstance?
			dynamic_cast<SurfaceSearchable*>(pInstance):
			new SurfaceSearchable();

	checkCache();

	SurfaceSphere::clone(pSurfaceSearchable);

	// TODO copy attributes
	const U32 sizeVector=m_vertices*sizeof(SpatialVector);
	const U32 sizeI32=m_vertices*sizeof(I32);

	pSurfaceSearchable->m_vertices=m_vertices;

	pSurfaceSearchable->m_pVertexArray=(SpatialVector*)reallocate(
			pSurfaceSearchable->m_pVertexArray,sizeVector);
	memcpy(pSurfaceSearchable->m_pVertexArray,m_pVertexArray,sizeVector);

	pSurfaceSearchable->m_pNormalArray=(SpatialVector*)reallocate(
			pSurfaceSearchable->m_pNormalArray,sizeVector);
	memcpy(pSurfaceSearchable->m_pNormalArray,m_pNormalArray,sizeVector);

	pSurfaceSearchable->m_pPickIndexArray=(I32*)reallocate(
			pSurfaceSearchable->m_pPickIndexArray,sizeI32);
	memcpy(pSurfaceSearchable->m_pPickIndexArray,m_pNormalArray,sizeI32);

	return pSurfaceSearchable;
}

void SurfaceSearchable::cache(void)
{
	feLog("SurfaceSearchable::cache\n");

	if(!m_record.isValid())
	{
		feLog("SurfaceSearchable::cache record invalid\n");
		return;
	}

	SurfaceSphere::cache();

	SurfaceModel surfaceModelRV;
	surfaceModelRV.bind(m_record);

	sp<RecordGroup> spRG=surfaceModelRV.surfacePointRG();
	for(RecordGroup::iterator it=spRG->begin();it!=spRG->end();it++)
	{
		sp<RecordArray> spRA= *it;

		SurfacePoint m_surfacePointRV;
		m_surfacePointRV.bind(spRA);
		while(m_surfacePointRV.step())
		{
			SpatialVector& location=m_surfacePointRV.location();
			SpatialVector& up=m_surfacePointRV.up();
//			SpatialVector& texCoord=m_surfacePointRV.texCoord();

			m_pVertexArray=(SpatialVector*)reallocate(m_pVertexArray,
					(m_vertices+1)*sizeof(SpatialVector));
			m_pNormalArray=(SpatialVector*)reallocate(m_pNormalArray,
					(m_vertices+1)*sizeof(SpatialVector));
			m_pPickIndexArray=(I32*)reallocate(m_pPickIndexArray,
					(m_vertices+1)*sizeof(I32));

			SpatialVector& rVertex=m_pVertexArray[m_vertices];
			SpatialVector& rNormal=m_pNormalArray[m_vertices];
			I32& rPickIndex=m_pPickIndexArray[m_vertices];

			rVertex=location;
			rNormal=up;

			printf("vertex %d  %s  %s  %d\n",m_vertices,
					c_print(rVertex),c_print(rNormal),rPickIndex);
			m_vertices++;
		}
	}

	if(surfaceModelRV.searchComponent().isValid())
	{
		m_spSearchComponent=surfaceModelRV.searchComponent();
	}
}

void SurfaceSearchable::prepareForSearch(void)
{
	checkCache();

	if(m_spSpatialTreeI.isValid())
	{
		return;
	}

#if	FE_SS_DEBUG
	feLog("SurfaceSearchable::prepareForSearch m_vertices=%d tree valid=%d\n",
			m_vertices,m_spSpatialTreeI.isValid());
#endif

	if(!m_vertices)
	{
		return;
	}
	if(m_spSearchComponent.isValid())
	{
		m_spSpatialTreeI=m_spSearchComponent;
	}
	if(!m_spSpatialTreeI.isValid())
	{
		m_spSpatialTreeI=create("SpatialTreeI.default");
	}
	if(!m_spSpatialTreeI.isValid())
	{
		return;
	}

#if	FE_SS_DEBUG
	feLog("SurfaceSearchable::prepareForSearch building spatial tree\n");
#endif

	m_spSpatialTreeI->setRefinement(m_refinement);
	m_spSpatialTreeI->populate(m_pVertexArray,m_pNormalArray,m_pPickIndexArray,
			m_vertices,center(),radius());
}

SpatialTransform SurfaceSearchable::sample(Vector2 a_uv) const
{
	SpatialTransform result;
	setIdentity(result);
	return result;
}

void SurfaceSearchable::Impact::copy(sp<SpatialTreeI::Hit>& a_rspHit)
{
	setPickIndex(a_rspHit->pickIndex());
	setDirection(a_rspHit->direction());
	setDistance(a_rspHit->distance());
}

std::vector< sp<SurfaceI::ImpactI> > SurfaceSearchable::nearestPoints(
		const SpatialVector& a_origin,U32 a_hitLimit) const
{
	std::vector< sp<SurfaceI::ImpactI> > impactArray;

	if(!m_vertices || !m_spSpatialTreeI.isValid())
	{
		return impactArray;
	}

	std::vector< sp<SpatialTreeI::Hit> > hitArray;
	m_spSpatialTreeI->nearestPoint(a_origin,a_hitLimit,hitArray);
	const U32 hitCount=hitArray.size();
	if(!hitCount)
	{
		feLog("SurfaceSearchable::nearestPoints no hits\n");
		return impactArray;
	}

	impactArray.resize(hitCount);

	for(U32 m=0;m<hitCount;m++)
	{
		sp<SpatialTreeI::Hit>& rspHit=hitArray[m];
		sp<Impact> spImpact=m_pImpactPool->get();
		spImpact->copy(rspHit);
		spImpact->setSurface(dynamic_cast<const SurfaceSearchable*>(this));
		spImpact->setOrigin(a_origin);
		impactArray[m]=spImpact;
	}

	return impactArray;
}

sp<SurfaceSearchable::ImpactI> SurfaceSearchable::nearestPoint(
		const SpatialVector& a_origin) const
{
	if(!m_vertices || !m_spSpatialTreeI.isValid())
	{
		return sp<ImpactI>(NULL);
	}

	std::vector< sp<SpatialTreeI::Hit> > hitArray;
	const U32 hitLimit=1;	//* TODO param multi-weighting
	m_spSpatialTreeI->nearestPoint(a_origin,hitLimit,hitArray);
	const U32 hitCount=hitArray.size();
	if(!hitCount)
	{
		feLog("SurfaceSearchable::nearestPoint no hits\n");
		return sp<ImpactI>(NULL);
	}

#if FE_SS_TREE_HIT_DEBUG
	feLog("\n");
	for(U32 m=0;m<hitCount;m++)
	{
		sp<SpatialTreeI::Hit>& rspHit=hitArray[m];
		feLog("Hit %d/%d dist %.6G  inter %s\n", m,hitCount,
				rspHit->distance(),c_print(rspHit->intersection()));
		feLog("  face %d bary %s\n",rspHit->face(),
				c_print(rspHit->barycenter()));
		for(U32 n=0;n<3;n++)
		{
			feLog("  v%d %s n %s\n",n,c_print(rspHit->vertex()[n]),
					c_print(rspHit->normal()[n]));
/*
			FEASSERT(fabs(magnitude(rspHit->vertex()[n]))<1e9);
			FEASSERT(fabs(magnitude(rspHit->normal()[n]))<1.1);
			if(fabs(magnitude(rspHit->vertex()[n]))>1e9 ||
					fabs(magnitude(rspHit->normal()[n]))>1.1)
			{
				exit(1);
			}
*/
		}
	}
#endif

	Real hitWeight[hitLimit];
	hitWeight[0]=1.0;

#if TRUE
	//* simply closest
	const Real sum=1.0;
	for(U32 m=1;m<hitCount;m++)
	{
		hitWeight[m]=0.0;
	}
#endif

	// TODO pick single best hit
	U32 bestHit=0;
	Real bestScore=0.0;
	for(U32 m=0;m<hitCount;m++)
	{
		const sp<SpatialTreeI::Hit>& rspHit=hitArray[m];
		Real thisScore=0.0;
		const SpatialVector diff=a_origin-rspHit->intersection();
		const Real mag=magnitude(diff);
		if(mag<1e-6)
		{
			thisScore=1.0;
		}
		else
		{
			const SpatialVector to=diff/mag;
			const SpatialVector norm=unitSafe(location(rspHit->barycenter(),
					rspHit->normal()[0],rspHit->normal()[1],
					rspHit->normal()[2]));
			const Real to_dot=dot(to,norm);
			thisScore=(to_dot>0.0)? to_dot: 0.0;
#if FE_SS_TREE_HIT_DEBUG
			feLog("  to %s\n",c_print(to));
#endif
		}
		if(!m)
		{
			bestScore=thisScore;
		}
		else if(thisScore>bestScore)
		{
			bestScore=thisScore;
			bestHit=m;
		}
#if FE_SS_TREE_HIT_DEBUG
		feLog("Score %d/%d %.6G best %.6G\n",
				m,hitCount,thisScore,bestScore);
#endif
	}
	//* single weighted
	for(U32 m=0;m<hitCount;m++)
	{
		hitWeight[m]=0.0;
	}
	hitWeight[bestHit]=1.0;

	sp<SpatialTreeI::Hit>& rspHit=hitArray[bestHit];
	sp<Impact> spImpact=m_pImpactPool->get();
	spImpact->copy(rspHit);
	spImpact->setSurface(dynamic_cast<const SurfaceSearchable*>(this));
	spImpact->setOrigin(a_origin);

#if FALSE
	//* reevaluate barycenter based on normals
//	const SpatialVector to=unit(a_origin-rspHit->intersection());
	const SpatialVector norm=unitSafe(location(rspHit->barycenter(),
			rspHit->normal()[0],rspHit->normal()[1],rspHit->normal()[2]));
	SpatialBary barynormal;
	barynormal.solve(rspHit->normal()[0],rspHit->normal()[1],
			rspHit->normal()[2],norm);
	spImpact->setBarycenter(barynormal);
	const SpatialVector intersection=location(barynormal,
			rspHit->vertex()[0],rspHit->vertex()[1],rspHit->vertex()[2]);
	const SpatialVector toward=intersection-a_origin;
	const Real range=magnitude(toward);
	spImpact->setDistance(range);
	spImpact->setDirection(range>0.0? toward/range: toward);
#endif

	if(hitCount>1)
	{
		SpatialVector intersection;
		SpatialVector normal;
		set(intersection);
		set(normal);
		for(U32 m=0;m<hitCount;m++)
		{
			const Real fraction=hitWeight[m]/sum;
			intersection+=fraction*hitArray[m]->intersection();

			const SpatialVector* norms=hitArray[m]->normal();
			const Vector2& bary=hitArray[m]->barycenter();
			normal+=fraction*(norms[0]*bary[0]+norms[1]*bary[1]+
					norms[2]*(1.0f-bary[0]-bary[1]));
		}
		normalizeSafe(normal);
		const SpatialVector to=intersection-a_origin;
		const Real mag=magnitude(to);
		spImpact->setDistance(mag);
		spImpact->setDirection(mag>1e-6? to/mag: to);

#if FE_SS_TREE_HIT_DEBUG
		feLog("Blend %.6G %s\n",
				spImpact->distance(),c_print(intersection));
#endif
	}

/*
	feLog("dist %.6G dir %s\n",spImpact->distance(),
			c_print(spImpact->direction()));
	feLog("normal0 %s\n",c_print(spImpact->normal0()));
	feLog("normal1 %s\n",c_print(spImpact->normal1()));
	feLog("normal2 %s\n",c_print(spImpact->normal2()));
*/

	return spImpact;
}

std::vector< sp<SurfaceI::ImpactI> > SurfaceSearchable::rayImpacts(
		const SpatialVector& a_origin,const SpatialVector& a_direction,
		Real a_maxDistance,U32 a_hitLimit) const
{
	std::vector< sp<SurfaceI::ImpactI> > impactArray;

	if(!m_vertices || !m_spSpatialTreeI.isValid())
	{
		return impactArray;
	}

	std::vector< sp<SpatialTreeI::Hit> > hitArray;
	const Real distance=m_spSpatialTreeI->rayImpact(a_origin,a_direction,
			a_maxDistance,a_hitLimit,hitArray);
	const U32 hitCount=hitArray.size();

	if(!hitCount || (a_maxDistance>0.0 && distance>a_maxDistance))
	{
		return impactArray;
	}

	impactArray.resize(hitCount);
	for(U32 m=0;m<hitCount;m++)
	{
		sp<SpatialTreeI::Hit>& rspHit=hitArray[m];
		sp<Impact> spImpact=m_pImpactPool->get();
		spImpact->copy(rspHit);
		spImpact->setSurface(dynamic_cast<const SurfaceSearchable*>(this));
		spImpact->setOrigin(a_origin);
		spImpact->setDirection(a_direction);
		impactArray[m]=spImpact;
	}
	return impactArray;
}

sp<SurfaceSearchable::ImpactI> SurfaceSearchable::rayImpact(
		const SpatialVector& a_origin,const SpatialVector& a_direction,
		Real a_maxDistance) const
{
	if(!m_vertices || !m_spSpatialTreeI.isValid())
	{
		return sp<ImpactI>(NULL);
	}

	const U32 hitLimit=1;
	std::vector< sp<SpatialTreeI::Hit> > hitArray;
	const Real distance=m_spSpatialTreeI->rayImpact(a_origin,a_direction,
			a_maxDistance,hitLimit,hitArray);
	const U32 hitCount=hitArray.size();
	if(!hitCount || (a_maxDistance>0.0 && distance>a_maxDistance))
	{
		return sp<ImpactI>(NULL);
	}

	sp<SpatialTreeI::Hit>& rspHit=hitArray[0];

	sp<Impact> spImpact=m_pImpactPool->get();
	spImpact->copy(rspHit);
	spImpact->setSurface(dynamic_cast<const SurfaceSearchable*>(this));
	spImpact->setOrigin(a_origin);
	spImpact->setDirection(a_direction);
	return spImpact;
}

void SurfaceSearchable::calcBoundingSphere(void)
{
	if(!m_vertices)
	{
		return;
	}

#if FALSE
	//* by center of "mass"
	m_center=m_pVertexArray[0];
	for(U32 m=1;m<m_vertices;m++)
	{
		m_center+=m_pVertexArray[m];
	}
	m_center*=1.0/m_vertices;
#else
	//* by bounding box
	SpatialVector min=m_pVertexArray[0];
	SpatialVector max=min;
	for(U32 m=1;m<m_vertices;m++)
	{
		SpatialVector& rVertex=m_pVertexArray[m];
		for(U32 n=0;n<3;n++)
		{
			if(min[n]>rVertex[n])
			{
				min[n]=rVertex[n];
			}
			if(max[n]<rVertex[n])
			{
				max[n]=rVertex[n];
			}
		}
	}
	m_center=0.5*(min+max);
#endif

	Real max2=0.0;
	for(U32 m=1;m<m_vertices;m++)
	{
		SpatialVector radial=m_pVertexArray[m]-m_center;
		Real dist2=magnitudeSquared(radial);
		if(max2<dist2)
		{
			max2=dist2;
		}
	}
	m_radius=sqrt(max2);

#if FE_SS_DEBUG
	feLog("SurfaceSearchable::calcBoundingSphere"
			" vertices %d center %s radius %.6G\n",
			m_vertices,c_print(m_center),m_radius);
#endif
}

} // namespace
